# Authors : Florian Lalande < florianlalande@orange.fr >
#           Austin Peel < austin.peel@cea.fr >

import os
import sys
import time
import numpy as np
import argparse
from argparse import RawTextHelpFormatter
from keras.models import Sequential
from keras.layers import Dense, Dropout, Flatten
from keras.layers.convolutional import Conv3D
from keras.layers.convolutional import MaxPooling3D
from keras.optimizers import Adam
from keras.utils import np_utils
from keras import backend as K
from keras import callbacks
K.set_image_dim_ordering('th')
from sklearn.metrics import confusion_matrix

# Only use a specific GPU
# os.environ["CUDA_VISIBLE_DEVICES"] = "1"


def save_time(filepath, text=''):
    """Save current time to file."""
    hour = time.localtime().tm_hour
    min = time.localtime().tm_min
    sec = time.localtime().tm_sec
    with open(filepath, 'w') as f:
        f.write(text + "{:02}:{:02}:{:02}\n".format(hour, min, sec))


class History(callbacks.Callback):
    """Keras class used to track performance while training."""
    def on_train_begin(self, logs={}):
        self.losses = []
        self.valacc = []

    def on_epoch_end(self, batch, logs={}):
        self.losses.append(logs.get('loss'))
        self.valacc.append(logs.get('val_acc'))


class mgcnn(object):
    def __init__(self):
        self.model = self.create_model(verbose=True)

    def create_model(self, verbose=False):
        model = Sequential()
        model.add(Conv3D(8, (2, 3, 10), input_shape=(1, 4, 5, 100),
                         padding="same", activation='relu'))
        model.add(Conv3D(8, (2, 3, 10), padding="same", activation='relu'))
        model.add(MaxPooling3D(pool_size=(1, 1, 5)))
        model.add(Conv3D(8, (2, 3, 10), padding="same", activation='relu'))
        model.add(MaxPooling3D(pool_size=(1, 1, 2)))
        model.add(Dropout(0.3))
        model.add(Flatten())
        model.add(Dense(32, activation='relu'))
        model.add(Dense(16, activation='relu'))
        model.add(Dense(4, activation='softmax'))
        # Compile model
        optimizer = Adam(lr=0.001, decay=0.001) # lr <- lr * (1.0/(1.0+decay*iter))
        model.compile(loss='categorical_crossentropy', optimizer=optimizer,
                      metrics=['accuracy'])
        if verbose:
            model.summary()
        return model

    def load_data(self, datapath):
        try:
            self.data = np.load(datapath)
        except Exception:
            print("Unable to load " + datapath)

    def reset(self, verbose=False):
        # Generate a new model for independent training
        del self.model
        self.model = self.create_model(verbose)
        if verbose:
            print("Resetting CNN model.")

    def train(self, savepath, iterations, epochs):
        if not hasattr(self, 'data'):
            print("No data loaded.")
            return

        save_time(os.path.join(savepath, "je_rentre.txt"), "start time ")

        X = self.data
        X = X.reshape(1024, 1, 4, 5, 100)

        y = np.floor(np.linspace(0, 3.9999, 1024))
        y = np_utils.to_categorical(y)

        h = np.zeros((iterations, 2, epochs))

        for iteration in range(iterations):
            print("\nITERATION {}".format(iteration))
            print("-" * (10 + len(str(iteration))))
            # Randomize training and test data
            i1 = np.arange(256)
            i2 = np.arange(256, 512)
            i3 = np.arange(512, 768)
            i4 = np.arange(768, 1024)

            np.random.shuffle(i1)
            np.random.shuffle(i2)
            np.random.shuffle(i3)
            np.random.shuffle(i4)

            i_train = np.concatenate((i1[64:], i2[64:], i3[64:], i4[64:]))
            i_test = np.concatenate((i1[:64], i2[:64], i3[:64], i4[:64]))
            np.random.shuffle(i_train)
            np.random.shuffle(i_test)

            X_train = X[i_train]
            y_train = y[i_train]
            X_test = X[i_test]
            y_test = y[i_test]

            # Do training and save performance measures
            history = History()
            self.model.fit(X_train, y_train, validation_data=(X_test, y_test),
                           epochs=epochs, batch_size=200, verbose=1,
                           callbacks=[history])
            h[iteration, 0] = history.losses
            h[iteration, 1] = history.valacc
            cb_path = os.path.join(savepath, "callback{}.npy".format(iteration))
            np.save(cb_path, h[iteration])

            # Compute confusion matrix
            y_pred = self.model.predict(X_test)
            y_pred = np.argmax(y_pred, axis=1)
            y_test = np.argmax(y_test, axis=1)
            cm = confusion_matrix(y_test, y_pred) / 64.0
            cm_path = os.path.join(savepath, "cm{}.npy".format(iteration))
            np.save(cm_path, cm)

            # Save model
            model_json = self.model.to_json()
            model_name = "model{}.json".format(iteration)
            with open(os.path.join(savepath, "model", model_name), "w") as f:
                f.write(model_json)
            weights = "weights{}.hdf5".format(iteration)
            self.model.save_weights(os.path.join(savepath, "model", weights))

            # Start next iteration fresh
            self.reset()

        save_time(os.path.join(savepath, "je_sors.txt"), "end time ")


if __name__ == '__main__':
    description = ("Train a CNN to classify four cosmological models\n" +
                   "based on their weak-lensing convergence maps.")
    parser = argparse.ArgumentParser(description=description,
                                     formatter_class=RawTextHelpFormatter)
    parser.add_argument("-n", "--noise", type=int, default=0,
                        help=("noise level; options are\n" +
                              "0 for clean data\n" +
                              "1 for sigma = 0.35\n" +
                              "2 for sigma = 0.70\n" +
                              "(default is 0)"))
    parser.add_argument("-i", "--iterations", type=int, default=100,
                        help="number of iterations")
    parser.add_argument("-e", "--epochs", type=int, default=1000,
                        help="number of epochs")
    args = parser.parse_args()
    noise_opts = {0: "clean", 1: "sigma035", 2: "sigma070"}
    noise_level = noise_opts[args.noise]

    # Set data paths
    cwd = os.getcwd()
    datapath = os.path.join(cwd, "data", noise_level, "all_pdfs.npy")
    print("Training on {} data for {} epochs and {} iterations.".format(
          noise_level, args.epochs, args.iterations))
    if not os.path.isdir("output"):
        os.mkdir("output")
        os.mkdir(os.path.join("output", noise_level))
        os.mkdir(os.path.join("output", noise_level, "model"))
    outpath = os.path.join(cwd, "output", noise_level)

    # Build the CNN and train
    cnn = mgcnn()
    cnn.load_data(datapath)
    cnn.train(savepath=outpath, iterations=args.iterations, epochs=args.epochs)

    sys.exit()
